@@ -1,23 +1,26 @@
 /*
  * Copyright © 2008 Kristian Høgsberg
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #include <stddef.h>
@@ -33,7 +36,6 @@
 #include <sys/signalfd.h>
 #include <sys/timerfd.h>
 #include <unistd.h>
-#include <assert.h>
 #include "wayland-private.h"
 #include "wayland-server.h"
 #include "wayland-os.h"
@@ -97,7 +99,6 @@ add_source(struct wl_event_loop *loop,
 	struct epoll_event ep;
 
 	if (source->fd < 0) {
-		fprintf(stderr, "could not add source\n: %m");
 		free(source);
 		return NULL;
 	}
@@ -135,7 +136,7 @@ wl_event_loop_add_fd(struct wl_event_loop *loop,
 		return NULL;
 
 	source->base.interface = &fd_source_interface;
-	source->base.fd = fcntl(fd, F_DUPFD_CLOEXEC, 0);
+	source->base.fd = wl_os_dupfd_cloexec(fd, 0);
 	source->func = func;
 	source->fd = fd;
 
@@ -173,9 +174,9 @@ wl_event_source_timer_dispatch(struct wl_event_source *source,
 	int len;
 
 	len = read(source->fd, &expires, sizeof expires);
-	if (len != sizeof expires)
+	if (!(len == -1 && errno == EAGAIN) && len != sizeof expires)
 		/* Is there anything we can do here?  Will this ever happen? */
-		fprintf(stderr, "timerfd read error: %m\n");
+		wl_log("timerfd read error: %m\n");
 
 	return timer_source->func(timer_source->base.data);
 }
@@ -196,7 +197,8 @@ wl_event_loop_add_timer(struct wl_event_loop *loop,
 		return NULL;
 
 	source->base.interface = &timer_source_interface;
-	source->base.fd = timerfd_create(CLOCK_MONOTONIC, TFD_CLOEXEC);
+	source->base.fd = timerfd_create(CLOCK_MONOTONIC,
+					 TFD_CLOEXEC | TFD_NONBLOCK);
 	source->func = func;
 
 	return add_source(loop, &source->base, WL_EVENT_READABLE, data);
@@ -211,10 +213,8 @@ wl_event_source_timer_update(struct wl_event_source *source, int ms_delay)
 	its.it_interval.tv_nsec = 0;
 	its.it_value.tv_sec = ms_delay / 1000;
 	its.it_value.tv_nsec = (ms_delay % 1000) * 1000 * 1000;
-	if (timerfd_settime(source->fd, 0, &its, NULL) < 0) {
-		fprintf(stderr, "could not set timerfd\n: %m");
+	if (timerfd_settime(source->fd, 0, &its, NULL) < 0)
 		return -1;
-	}
 
 	return 0;
 }
@@ -235,9 +235,9 @@ wl_event_source_signal_dispatch(struct wl_event_source *source,
 	int len;
 
 	len = read(source->fd, &signal_info, sizeof signal_info);
-	if (len != sizeof signal_info)
+	if (!(len == -1 && errno == EAGAIN) && len != sizeof signal_info)
 		/* Is there anything we can do here?  Will this ever happen? */
-		fprintf(stderr, "signalfd read error: %m\n");
+		wl_log("signalfd read error: %m\n");
 
 	return signal_source->func(signal_source->signal_number,
 				   signal_source->base.data);
@@ -265,7 +265,7 @@ wl_event_loop_add_signal(struct wl_event_loop *loop,
 
 	sigemptyset(&mask);
 	sigaddset(&mask, signal_number);
-	source->base.fd = signalfd(-1, &mask, SFD_CLOEXEC);
+	source->base.fd = signalfd(-1, &mask, SFD_CLOEXEC | SFD_NONBLOCK);
 	sigprocmask(SIG_BLOCK, &mask, NULL);
 
 	source->func = func;
@@ -423,10 +423,12 @@ wl_event_loop_dispatch(struct wl_event_loop *loop, int timeout)
 
 	wl_event_loop_process_destroy_list(loop);
 
+	wl_event_loop_dispatch_idle(loop);
+
 	do {
 		n = post_dispatch_check(loop);
 	} while (n > 0);
-		
+
 	return 0;
 }
 
